import os
import sys
import datetime
import pytest
import pandas as pd


test_year = 2023


def setup():
    code_dir = os.path.join(os.path.abspath(".."), "src")
    sys.path.append(code_dir)
    from _pytest.assertion import truncate

    truncate.DEFAULT_MAX_LINES = 9999
    truncate.DEFAULT_MAX_CHARS = 9999


setup()

from markets_insights.core.column_definition import BaseColumns, BasePriceColumns
from markets_insights.datareader.data_reader import DataReader
from markets_insights.core.core import (
    DateFilter,
    DatePartsFilter,
    IdentifierFilter,
    TypeHelper,
    MarketDaysHelper,
)


class PresetDates:
    year_start: datetime.datetime = MarketDaysHelper.get_this_or_next_market_day(
        datetime.date(year=test_year, month=1, day=1)
    )
    year_end: datetime.datetime = MarketDaysHelper.get_this_or_previous_market_day(
        datetime.date(year=test_year, month=12, day=31)
    )
    dec_start: datetime.datetime = MarketDaysHelper.get_this_or_next_market_day(
        datetime.date(year=test_year, month=12, day=1)
    )
    dec_end: datetime.datetime = year_end
    q4_start: datetime.datetime = MarketDaysHelper.get_this_or_next_market_day(
        datetime.date(year=test_year, month=10, day=1)
    )
    q4_end: datetime.datetime = year_end
    recent_day: datetime.datetime = MarketDaysHelper.get_this_or_previous_market_day(
        datetime.date.today() - datetime.timedelta(days=1)
    )


class PresetFilters:
    nifty50 = IdentifierFilter("NIFTY 50")
    date_year_end = DateFilter(PresetDates.year_end)
    year = DatePartsFilter(year=2023)


class Presets:
    cols_to_test = [
        BaseColumns.Identifier,
        BaseColumns.Date,
        BaseColumns.Close,
        BaseColumns.Open,
        BaseColumns.High,
        BaseColumns.Turnover,
        BaseColumns.Volume,
    ]
    filters = PresetFilters
    dates = PresetDates


def check_cols_present(data: pd.DataFrame, cols_to_test: [int], case_identifier: str):
    for col_name in cols_to_test:
        if not col_name in data.columns:
            assert pytest.fail(
                f"column {col_name} not present in {case_identifier} data"
            )


def check_base_cols_present(data: pd.DataFrame, case_identifier: str):
    check_cols_present(data, Presets.cols_to_test, case_identifier)


def check_price_cols_present(data: pd.DataFrame, period: str):
    check_cols_present(
        data,
        [
            col_template.substitute({"period": period})
            for col_template in TypeHelper.get_class_static_values(BasePriceColumns)
        ],
        period,
    )


def check_col_values(data: pd.DataFrame, col_value_pairs: dict):
    for col_name in col_value_pairs:
        if isinstance(col_value_pairs[col_name], float):
            if not float(data[col_name].values[0]) == pytest.approx(
                col_value_pairs[col_name]
            ):
                assert pytest.fail(
                    f"column `{col_name}` value check: {data[col_name].values[0]} <> {col_value_pairs[col_name]}"
                )
        else:
            if not data[col_name].values[0] == col_value_pairs[col_name]:
                assert pytest.fail(
                    f"column `{col_name}` value check: {data[col_name].values[0]} <> {col_value_pairs[col_name]}"
                )



from markets_insights.core.core import Instrumentation
from markets_insights.core.column_definition import BaseColumns, CalculatedColumns, DerivativesBaseColumns
import pandas as pd
import pandas_ta as ta

class CalculationWindow:
    def __init__(self, trailing: int = 0, leading: int = 0, ):
        self.trailing = trailing
        self.leading = leading

    def load_from_list(windows: list):
        max_leading: int = 0
        max_trailing: int = 0

        for window in windows:
            max_leading = max(max_leading, window.leading)
            max_trailing = max(max_trailing, window.trailing)
        
        return CalculationWindow(max_trailing, max_leading)

class CalculationWorker:
    def __init__(self, **params):
        self._columns: list[str] = []
        self._params: dict = params

    def get_calculation_window(self) -> CalculationWindow:
        max_leading = 0
        max_trailing = 0
        for window in [window for window in ['time_window', 'N'] if window in self._params]:
            window_val = int(self._params[window])
            max_trailing = max(max_trailing, window_val)  # Assuming negative values should set max_trailing

        return CalculationWindow(max_trailing, max_leading)

    def get_group_cols(self, columns: list[str]) -> list[str]:
        group_columns: list[str] = [DerivativesBaseColumns.Identifier]
                
        group_columns.extend(group_col for group_col in 
                [
                    DerivativesBaseColumns.OptionType,
                    DerivativesBaseColumns.ExpiryDate,
                    DerivativesBaseColumns.StrikePrice,
                    DerivativesBaseColumns.OptionType,
                ] if group_col in columns
            )

        return group_columns

    def get_columns(self) -> list[str]:
        return self._columns
    
    def get_column(self) -> str:
        return self._columns[0]
    
    def get_params(self) -> dict:
        return self._params

    def add_calculated_columns(self, data: pd.DataFrame):
        raise NotImplementedError("add_calculated_fields")


class CalculationPipeline:
    _pipeline: list[CalculationWorker]

    def __init__(self, workers: list[CalculationWorker]=None):
        if workers is not None:
            self._pipeline = workers
        else:
            self._pipeline = []

    def add_calculation_worker(self, worker: CalculationWorker):
        self._pipeline.append(worker)

    def run(self, data: pd.DataFrame):
        for worker in self._pipeline:
            result = worker.add_calculated_columns(data)
            if result is not None:
                data = result
        return data
    
    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow.load_from_list([worker.get_calculation_window() for worker in self._pipeline])


class ColumnValueCrossedAboveFlagWorker(CalculationWorker):
    def __init__(self, value_column: str = None, value: int = 0):
        super().__init__(value_column = value_column, value = int(value))
        self._columns.append(f"{value_column}CrossedAbove")

    @Instrumentation.trace(name="ColumnValueCrossedAboveFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[self._columns[0]] = identifier_grouped_data[self._params['value_column']].transform(
            lambda x: (x.shift(1) < self._params['value']) & (x >= self._params['value'])
        )
    
    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow(trailing=1, leading=0)


class ColumnValueCrossedBelowFlagWorker(CalculationWorker):
    def __init__(self, value_column: str = None, value: int = 0):
        super().__init__(value_column = value_column, value = int(value))
        self._columns.append(f"{value_column}CrossedBelow")

    @Instrumentation.trace(name="ColumnValueCrossedBelowFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[self._columns[0]] = identifier_grouped_data[self._params['value_column']].transform(
            lambda x: (x.shift(-1) > self._params['value']) & (x <= self._params['value'])
        )

    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow(trailing=1, leading=0)
    

class PriceCrossedAboveColumnValueFlagWorker(CalculationWorker):
    def __init__(self, value_column: str = None):
        super().__init__(value_column = value_column)
        self._columns.append(f"PriceCrossedAbove{value_column}")

    @Instrumentation.trace(name="PriceCrossedAboveColumnValueFlagWorker")
    def add_calculated_columns(self, data):
        data[self._columns[0]] = (
            data[BaseColumns.Close] >= data[self._params['value_column']]
        ) & (data[BaseColumns.PreviousClose] < data[self._params['value_column']])


class PriceCrossedBelowColumnValueFlagWorker(CalculationWorker):
    def __init__(self, value_column: str = None):
        super().__init__(value_column = value_column)
        self._columns.append(f"PriceCrossedBelow{value_column}")

    @Instrumentation.trace(name="PriceCrossedBelowColumnValueFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[BaseColumns.Close] < data[self._params['value_column']]
        ) & (data[BaseColumns.PreviousClose] >= data[self._params['value_column']])


class ColumnValueCrossedAboveAnotherColumnValueFlagWorker(CalculationWorker):
    def __init__(self, value_column_a: str = None, value_column_b: str = None):
        super().__init__(value_column_a=value_column_a, value_column_b=value_column_b)
        self._columns.append(f"{value_column_a}CrossedAbove{value_column_b}")

    @Instrumentation.trace(name="ColumnValueCrossedAboveAnotherColumnValueFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        grouped_data = data.groupby(self.get_group_cols(data.columns))
        column_a = self._params["value_column_a"]
        column_b = self._params["value_column_b"]
        if len(data[BaseColumns.Identifier].unique()) > 1:
            data[self._columns[0]] = grouped_data.apply(
                lambda x: (x.shift(1)[column_a] < x.shift(1)[column_b])
                & (x[column_a] >= x[column_b])
            ).reset_index(level=0, drop=True)
        else:
            data[self._columns[0]] = (data.shift(1)[column_a] < data.shift(1)[column_b]) \
                & (data[column_a] >= data[column_b])
    
    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow(trailing=1, leading=0)
    

class ColumnValueCrossedBelowAnotherColumnValueFlagWorker(CalculationWorker):
    def __init__(self, value_column_a: str = None, value_column_b: str = None):
        super().__init__(value_column_a=value_column_a, value_column_b=value_column_b)
        self._columns.append(f"{value_column_a}CrossedBelow{value_column_b}")

    @Instrumentation.trace(name="ColumnValueCrossedBelowAnotherColumnValueFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        column_a = self._params["value_column_a"]
        column_b = self._params["value_column_b"]
        if len(data[BaseColumns.Identifier].unique()) > 1:
            data[self._columns[0]] = identifier_grouped_data.apply(
                lambda x: (x.shift(-1)[column_a] > x.shift(-1)[column_b])
                & (x[column_a] <= x[column_b])
            ).reset_index(level=0, drop=True)
        else:
            data[self._columns[0]] = (data.shift(-1)[column_a] > data.shift(-1)[column_b]) \
                & (data[column_a] <= data[column_b])

    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow(trailing=1, leading=0)
    

class ColumnValueBelowFlagWorker(CalculationWorker):
    def __init__(self, value_column: str = None, value: int = 0):
        super().__init__(value_column=value_column, value=int(value))
        self._columns.append(f"{value_column}Below{str(value)}")

    @Instrumentation.trace(name="ColumnValueBelowFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[self._params["value_column"]] < self._params["value"]
        )


class ColumnValueAboveFlagWorker(CalculationWorker):
    def __init__(self, value_column: str = None, value: int = 0):
        super().__init__(value_column=value_column, value=int(value))
        self._columns.append(f"{value_column}Above{str(value)}")

    @Instrumentation.trace(name="ColumnValueAboveFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[self._params["value_column"]] > self._params["value"]
        )



class ColumnValueBelowAnotherColumnValueFlagWorker(CalculationWorker):
    def __init__(self, value_column_a: str = None, value_column_b: str = None):
        super().__init__(value_column_a=value_column_a, value_column_b=value_column_b)
        self._columns.append(f"{value_column_a}Below{value_column_b}")

    @Instrumentation.trace(name="ColumnValueBelowAnotherColumnValueFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[self._params["value_column_a"]] < data[self._params["value_column_b"]]
        )


class ColumnValueAboveAnotherColumnValueFlagWorker(CalculationWorker):
    def __init__(self, value_column_a: str = None, value_column_b: str = None):
        super().__init__(value_column_a=value_column_a, value_column_b=value_column_b)
        self._columns.append(f"{value_column_a}Above{value_column_b}")

    @Instrumentation.trace(name="ColumnValueAboveAnotherColumnValueFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[self._params["value_column_a"]] > data[self._params["value_column_b"]]
        )

class DatePartsCalculationWorker(CalculationWorker):
    def __init__(self):
        super().__init__()
        self._columns.append(CalculatedColumns.Year)
        self._columns.append(CalculatedColumns.MonthNo)
        self._columns.append(CalculatedColumns.Month)
        self._columns.append(CalculatedColumns.Day)

    @Instrumentation.trace(name="DatePartsCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[CalculatedColumns.Year] = data[BaseColumns.Date].dt.year
        data[CalculatedColumns.MonthNo] = data[BaseColumns.Date].dt.month
        data[CalculatedColumns.Month] = data[BaseColumns.Date].dt.strftime("%b")
        data[CalculatedColumns.Day] = data[BaseColumns.Date].dt.strftime("%A")


class SmaCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 50):
        super().__init__(time_window = int(time_window))
        self._columns.append(f"Sma{str(time_window)}")

    @Instrumentation.trace(name="SmaCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[self._columns[0]] = identifier_grouped_data[BaseColumns.Close].transform(
            lambda x: x.rolling(self._params['time_window']).mean()
        )


class StdDevCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 200):
        super().__init__(time_window = int(time_window))
        self._columns.append(f"StdDev{time_window}")

    @Instrumentation.trace(name="StdDevCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[self._columns[0]] = identifier_grouped_data[BaseColumns.Close].transform(
            lambda x: x.rolling(self._params['time_window']).std()
        )


class BollingerBandCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 200, deviation: int = 2):
        super().__init__(time_window = int(time_window), deviation = int(deviation))
        self._columns.append(f"Bb{str(time_window)}Dev{str(deviation)}Lower")
        self._columns.append(f"Bb{str(time_window)}Dev{str(deviation)}Upper")

    @Instrumentation.trace(name="BollingerBandCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        sma_column = f"Sma{str(self._params['time_window'])}"
        std_dev_column = f"StdDev{str(self._params['time_window'])}"

        if sma_column in data.columns:
            SmaCalculationWorker(time_window=int(self._params['time_window'])).add_calculated_columns(data)
        
        if std_dev_column in data.columns:
            StdDevCalculationWorker(time_window=int(self._params['time_window'])).add_calculated_columns(data)

        data[self._columns[0]] = data[sma_column] - (
            data[std_dev_column] * self._params['deviation']
        )
        data[self._columns[1]] = data[sma_column] + (
            data[std_dev_column] * self._params['deviation']
        )


class RsiOldCalculationWorker(CalculationWorker):
    def __init__(self, time_window):
        super().__init__()
        self._time_window = time_window
        self._column_name = "Rsi" + str(self._time_window)

    def calculate_wsm_average(self, raw_data, data, avg_col_name, abs_col_name):
        step = 1
        for i, row in enumerate(data[avg_col_name].iloc[self._time_window + step :]):
            raw_data.at[data.index[i + self._time_window + step], avg_col_name] = (
                data[avg_col_name].iloc[i + self._time_window]
                * (self._time_window - step)
                + data[abs_col_name].iloc[i + self._time_window + step]
            ) / self._time_window

    @Instrumentation.trace(name="RsiOldCalculationWorker")
    def add_calculated_columns(self, data):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[CalculatedColumns.ClosePriceDiff] = identifier_grouped_data[
            BaseColumns.Close
        ].transform(lambda x: x.diff(1))
        data[CalculatedColumns.Gain] = identifier_grouped_data[
            CalculatedColumns.ClosePriceDiff
        ].transform(lambda x: x.clip(lower=0).round(2))
        data[CalculatedColumns.Loss] = identifier_grouped_data[
            CalculatedColumns.ClosePriceDiff
        ].transform(lambda x: x.clip(upper=0).abs().round(2))

        data[CalculatedColumns.ClosePriceDiff].fillna(0)
        data[CalculatedColumns.Gain].fillna(0)
        data[CalculatedColumns.Loss].fillna(0)

        # Get initial Averages
        data[CalculatedColumns.AvgGain] = identifier_grouped_data[
            CalculatedColumns.Gain
        ].transform(
            lambda x: x.rolling(
                window=self._time_window, min_periods=self._time_window
            ).mean()
        )
        data[CalculatedColumns.AvgLoss] = identifier_grouped_data[
            CalculatedColumns.Loss
        ].transform(
            lambda x: x.rolling(
                window=self._time_window, min_periods=self._time_window
            ).mean()
        )

        for identifier in data[BaseColumns.Identifier].unique():
            self.calculate_wsm_average(
                data,
                identifier_grouped_data.get_group(identifier),
                CalculatedColumns.AvgGain,
                CalculatedColumns.Gain,
            )
            self.calculate_wsm_average(
                data,
                identifier_grouped_data.get_group(identifier),
                CalculatedColumns.AvgLoss,
                CalculatedColumns.Loss,
            )

        data[CalculatedColumns.RelativeStrength] = (
            data[CalculatedColumns.AvgGain] / data[CalculatedColumns.AvgLoss]
        )
        data[CalculatedColumns.RelativeStrengthIndex] = 100 - (
            100 / (1.0 + data[CalculatedColumns.RelativeStrength])
        )


class RsiCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 14):
        super().__init__(time_window = int(time_window))
        self._columns.append(CalculatedColumns.RelativeStrengthIndex)

    def calculate_rsi(self, group):
        group[CalculatedColumns.RelativeStrengthIndex] = ta.rsi(
            group[BaseColumns.Close]
        )
        return group

    @Instrumentation.trace(name="RsiCalculationWorker")
    def add_calculated_columns(self, data):
        result = data.groupby(self.get_group_cols(data.columns), group_keys=True).apply(
            self.calculate_rsi
        )
        return result.reset_index(drop=True)


class StochRsiCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 14):
        super().__init__(time_window = int(time_window))
        self._columns.append(CalculatedColumns.StochRsi_K)
        self._columns.append(CalculatedColumns.StochRsi_D)

    def calculate_stoch_rsi(self, group: pd.DataFrame):
        window = self._params['time_window']
        data = ta.stochrsi(group[BaseColumns.Close], length=window, rsi_length=window, k=3, d=3)
        if data is not None:
            group[CalculatedColumns.StochRsi_K] = data[f"STOCHRSIk_{window}_{window}_3_3"]
            group[CalculatedColumns.StochRsi_D] = data[f"STOCHRSId_{window}_{window}_3_3"]
        return group

    @Instrumentation.trace(name="StochRsiCalculationWorker")
    def add_calculated_columns(self, data):
        result = data.groupby(self.get_group_cols(data.columns), group_keys=True).apply(
            self.calculate_stoch_rsi
        )
        return result.reset_index(drop=True)

class VwapCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 1):
        super().__init__(time_window = int(time_window))
        self._columns.append(CalculatedColumns.Vwap)

    @Instrumentation.trace(name="Vwap2CalculationWorker")
    def add_calculated_columns(self, data):
        data[BaseColumns.Turnover] = data[BaseColumns.Turnover].replace("-", 0)
        data[BaseColumns.Volume] = data[BaseColumns.Volume].replace("-", 0)
        if len(data[BaseColumns.Identifier].unique()) > 1:
            data[CalculatedColumns.Vwap] = (
                data.groupby(self.get_group_cols(data.columns))
                .apply(
                    lambda x: x[BaseColumns.Turnover].rolling(self._params['time_window']).sum()
                    / x.rolling(self._params['time_window'])[BaseColumns.Volume].sum()
                )
                .reset_index(level=0, drop=True)
            )
        else:
            data[CalculatedColumns.Vwap] = data[BaseColumns.Turnover].rolling(self._params['time_window']).sum() / data[BaseColumns.Volume].rolling(self._params['time_window']).sum()

class LowestPriceInNextNDaysCalculationWorker(CalculationWorker):
    def __init__(self, N: int = 5, close_price_column: str = BaseColumns.Close, low_price_column: str = BaseColumns.Low):
        super().__init__(N = int(N), close_price_column = close_price_column, low_price_column = low_price_column)
        self._columns.append(f"TroughInNext{str(N)}Sessions")
        self._columns.append(f"TroughPercInNext{str(N)}Sessions")

    @Instrumentation.trace(name="LowestPriceInNextNDaysCalculationWorker")
    def add_calculated_columns(self, data):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[self._columns[0]] = identifier_grouped_data[
            self._params['low_price_column']
        ].transform(
            lambda x: x.rolling(self._params['N']).min().shift(-self._params['N'])
        )
        data[self._columns[1]] = (
            (data[self._params['close_price_column']] - data[self._columns[0]])
            / data[self._params['close_price_column']]
            * 100
        )

    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow(trailing=0, leading=int(self._params['N']))

class HighestPriceInNextNDaysCalculationWorker(CalculationWorker):
    def __init__(self, N: int = 5, close_price_column: str = BaseColumns.Close, high_price_column: str = BaseColumns.High):
        super().__init__(N = int(N), close_price_column = close_price_column, high_price_column = high_price_column)
        self._columns.append(f"PeakInNext{str(N)}Sessions")
        self._columns.append(f"PeakPercInNext{str(N)}Sessions")

    @Instrumentation.trace(name="HighestPriceInNextNDaysCalculationWorker")
    def add_calculated_columns(self, data):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        data[self._columns[0]] = identifier_grouped_data[
            self._params['high_price_column']
        ].transform(
            lambda x: x.rolling(self._params['N']).max().shift(-self._params['N'])
        )
        data[self._columns[1]] = (
            (data[self._columns[0]] - data[self._params['close_price_column']])
            / data[self._params['close_price_column']]
            * 100
        )

    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow(trailing=0, leading=int(self._params['N']))


class ColumnChangeOverNDaysCalculationWorker(CalculationWorker):
    def __init__(self, value_column: str = BaseColumns.Close, N: int = 0):
        super().__init__(value_column=value_column, N=int(N))
        self._columns.append(f"{value_column}Change{str(N)}Sessions")
        self._columns.append(f"{value_column}ChangePerc{str(N)}Sessions")

    @Instrumentation.trace(name="ColumnChangeOverNDaysCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(self.get_group_cols(data.columns))
        N: int = self._params['N']
        data[self._columns[0]] = identifier_grouped_data[self._params["value_column"]].transform(lambda x:
            x - x.shift(N)
        )
        data[self._columns[1]] = (
            data[self._columns[0]] / (data[self._params["value_column"]] - data[self._columns[0]]) * 100
        )


class ColumnsDeltaCalculationWorker(CalculationWorker):
    def __init__(self, value_column_a: str = None, value_column_b: str = None):
        super().__init__(value_column_a=value_column_a, value_column_b=value_column_b)
        self._columns.append(f"{value_column_a}-{value_column_b}")
        self._columns.append(f"{value_column_a}-{value_column_b}AsPerc")

    @Instrumentation.trace(name="ColumnsDeltaCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[self._params["value_column_a"]] - data[self._params["value_column_b"]]
        )
        data[self._columns[1]] = (
            data[self._columns[0]] / data[self._params["value_column_a"]] * 100
        )


import pytest
import pandas as pd

from helper import (
    check_col_values,
    setup,
    check_cols_present,
    check_base_cols_present,
    PresetDates,
)
from markets_insights.core.column_definition import BaseColumns, DerivativesBaseColumns
from markets_insights.core.core import (
    IdentifierFilter,
    InstrumentTypeFilter,
    TypeHelper,
)

setup()

import markets_insights

from markets_insights.datareader.data_reader import (
    DataReader,
    BhavCopyReader,
    NseEquityFuturesDataReader,
    NseIndexFuturesDataReader,
    NseIndicesReader,
    NseDerivatiesReader,
    MultiDatesDataReader,
    DateRangeDataReader,
    ForDateCriteria,
    MultiDatesCriteria,
    DateRangeCriteria,
)


@pytest.mark.parametrize(
    "reader,rows",
    [
        (BhavCopyReader(), 1797),
        (NseIndicesReader(), 107),
        (NseDerivatiesReader(), 12450),
        (NseEquityFuturesDataReader(), 503),
        (NseIndexFuturesDataReader(), 12)
    ],
)
def test_single_day_reader(reader: DataReader, rows: int):
    data = reader.read(ForDateCriteria(PresetDates.dec_start))
    check_base_cols_present(data, reader.name)
    assert data.shape[0] == rows


@pytest.mark.parametrize(
    "reader", [BhavCopyReader(), NseIndicesReader(), NseDerivatiesReader(), NseEquityFuturesDataReader(), NseIndexFuturesDataReader()]
)
def test_recent_day_read(reader: DataReader):
    data = reader.read(ForDateCriteria(PresetDates.recent_day))
    check_base_cols_present(data, reader.name)
    assert data.shape[0] > 1


@pytest.mark.parametrize(
    "reader", [NseDerivatiesReader(), NseEquityFuturesDataReader(), NseIndexFuturesDataReader()]
)
def test_derivatives_data_reader_columns(reader: DataReader):
    data = reader.read(ForDateCriteria(PresetDates.dec_start))
    check_cols_present(
        data,
        [
            col_name
            for col_name in TypeHelper.get_class_static_values(DerivativesBaseColumns)
        ],
        "derivatives",
    )


def test_multi_dates_reader():
    reader = MultiDatesDataReader(NseIndicesReader())
    data = reader.read(
        MultiDatesCriteria([PresetDates.year_start, PresetDates.dec_end, PresetDates.dec_start])
    )
    check_base_cols_present(data, reader)
    assert data.shape[0] == 319


def test_date_range_reader():
    reader = DateRangeDataReader(NseIndicesReader())
    data = reader.read(DateRangeCriteria(from_date=PresetDates.dec_start, to_date=PresetDates.dec_end))
    check_base_cols_present(data, reader)
    assert data.shape[0] == 2146


@pytest.mark.parametrize(
    "reader,turnover",
    [(BhavCopyReader(), 19493616.1), (NseIndicesReader(), 224615300000.0)],
)
def test_turnover_rescaled(reader: DataReader, turnover: float):
    data = reader.read(ForDateCriteria(PresetDates.dec_start))
    check_col_values(data=data, col_value_pairs={BaseColumns.Turnover: turnover})


@pytest.mark.parametrize(
    "operation,close",
    [
        (lambda x, y: x + y, 4801.4),
        (lambda x, y: x - y, 12.799999999999727),
        (lambda x, y: x / y, 1.0053460301549513),
        (lambda x, y: x * y, 5763319.53),
    ],
)
def test_arithmatic_op(operation, close):
    l_reader = NseDerivatiesReader().set_filter(InstrumentTypeFilter("FUTSTK"))
    r_reader = BhavCopyReader()
    op_reader = operation(l_reader, r_reader)
    data = op_reader.read(ForDateCriteria(for_date=PresetDates.dec_start)).query(
        str(IdentifierFilter("RELIANCE"))
    )
    check_col_values(
        data,
        col_value_pairs={BaseColumns.Identifier: "RELIANCE", BaseColumns.Close: close},
    )
    assert f"{l_reader.col_prefix}{BaseColumns.Close}" in data.columns
    assert f"{r_reader.col_prefix}{BaseColumns.Close}" in data.columns


@pytest.mark.parametrize(
    "operation,symbol,close",
    [
        (lambda x, y: x + y, "+", 20280.280000000002),
        (lambda x, y: x - y, "-", 20255.52),
        (lambda x, y: x / y, "/", 1637.1486268174476),
        (lambda x, y: x * y, "*", 250916.60200000004),
    ],
)
def test_arithmatic_op_single_id_right(operation, symbol, close):
    indices_reader = NseIndicesReader()
    vix_reader = NseIndicesReader().set_filter(IdentifierFilter("India VIX"))
    op_reader = operation(indices_reader, vix_reader)
    data = op_reader.read(ForDateCriteria(for_date=PresetDates.dec_start)).query(
        str(IdentifierFilter(f"Nifty 50 {symbol} India VIX"))
    )

    check_col_values(data, col_value_pairs={BaseColumns.Close: close})

    assert op_reader.l_prefix == "Index-"
    assert op_reader.r_prefix == "India VIX-"

    assert f"{indices_reader.col_prefix}{BaseColumns.Close}" in data.columns
    assert f"India VIX-{BaseColumns.Close}" in data.columns


@pytest.mark.parametrize(
    "operation,symbol,close",
    [
        (lambda x, y: x + y, "+", 20280.280000000002),
        (lambda x, y: x - y, "-", -20255.52),
        (lambda x, y: x / y, "/", 0.0006108180916621851),
        (lambda x, y: x * y, "*", 250916.60200000004),
    ],
)
def test_arithmatic_op_single_id_left(operation, symbol, close):
    indices_reader = NseIndicesReader()
    vix_reader = NseIndicesReader().set_filter(IdentifierFilter("India VIX"))
    op_reader: DataReader = operation(vix_reader, indices_reader)
    data = op_reader.read(ForDateCriteria(for_date=PresetDates.dec_start)).query(
        str(IdentifierFilter(f"India VIX {symbol} Nifty 50"))
    )

    check_col_values(
        data,
        col_value_pairs={BaseColumns.Close: close},
    )

    assert op_reader.l_prefix == "India VIX-"
    assert op_reader.r_prefix == "Index-"

    assert f"{indices_reader.col_prefix}{BaseColumns.Close}" in data.columns
    assert f"India VIX-{BaseColumns.Close}" in data.columns


@pytest.mark.parametrize(
    "operation,symbol,close",
    [
        (lambda x, y: x + y, "+", 20280.280000000002),
        (lambda x, y: x - y, "-", 20255.52),
        (lambda x, y: x / y, "/", 1637.1486268174476),
        (lambda x, y: x * y, "*", 250916.60200000004),
    ],
)
def test_arithmatic_op_single_id_both(operation, symbol, close):
    indices_reader = NseIndicesReader().set_filter(IdentifierFilter("Nifty 50"))
    vix_reader = NseIndicesReader().set_filter(IdentifierFilter("India VIX"))
    op_reader: DataReader = operation(indices_reader, vix_reader)
    data = op_reader.read(ForDateCriteria(for_date=PresetDates.dec_start))

    check_col_values(
        data,
        col_value_pairs={BaseColumns.Identifier: f"Nifty 50 {symbol} India VIX", BaseColumns.Close: close},
    )

    assert op_reader.l_prefix == "Nifty 50-"
    assert op_reader.r_prefix == "India VIX-"

    assert f"Nifty 50-{BaseColumns.Close}" in data.columns
    assert f"India VIX-{BaseColumns.Close}" in data.columns


def test_arithmatic_multiple_ops():
    eq_reader = BhavCopyReader()
    fut_reader = NseDerivatiesReader().set_filter(InstrumentTypeFilter("FUTSTK"))
    op_reader = fut_reader - eq_reader
    data = (
        op_reader.read(ForDateCriteria(for_date=PresetDates.dec_start))
        .query(str(IdentifierFilter("RELIANCE")))
        .sort_values(DerivativesBaseColumns.ExpiryDate)
    )
    check_col_values(
        data,
        col_value_pairs={BaseColumns.Identifier: "RELIANCE", BaseColumns.Close: 29.0},
    )

    assert op_reader.l_prefix == "FO-"
    assert op_reader.r_prefix == "Cash-"

    assert f"{eq_reader.col_prefix}{BaseColumns.Close}" in data.columns
    assert f"{fut_reader.col_prefix}{BaseColumns.Close}" in data.columns
